import shutil
from ....framework import OnDemandDockerApi, ResourcePrerequisite, FileLike
from .controller import VideoToImagesController
from .settings import VideoToImagesSettings
from yo_fluq import Query


class VideoToImages(OnDemandDockerApi[VideoToImagesSettings, VideoToImagesController]):
    def __init__(self):
        pass

    def process(self, file_name: str, cap_result_count: int|None = None):
        config = self.controller.get_run_configuration(['--file', file_name])
        self.controller.get_deployment().stop().remove()
        self.controller.run_with_configuration(config)
        folder = self.controller.resource_folder('output')
        result = {}
        for file in Query.folder(folder):
            name = int(file.name.split('.')[0])
            fname = f'{self.current_job_id}.{name}.png'
            shutil.copy(file, self.cache_folder/fname)
            result[name] = fname
        result = [result[key] for key in sorted(result)]
        if cap_result_count is not None:
            result = result[:cap_result_count]
        return result


    @staticmethod
    def upload_video(file: FileLike.Type):
        return ResourcePrerequisite(
            VideoToImages,
            f'/input/{FileLike.get_name(file,True)}',
            file
        )

    Controller = VideoToImagesController
    Settings = VideoToImagesSettings


import base64
import os
import shutil
from io import BytesIO

import cv2
from loguru import logger

from fastapi import APIRouter, FastAPI
from fastapi.responses import JSONResponse

import numpy as np

from cv2 import cvtColor, COLOR_BGR2GRAY, Laplacian, CV_64F
from PIL import Image
from sentence_transformers import SentenceTransformer, util
from numpy.typing import NDArray
from model import get_comparator


class VideoProcessorApp:


    def __init__(self):
        self._similarity_threshold = 0.9
        self._blur_threshold = 150
        self.processed_frames = []
        self.start_index_batch = 0
        self.input_directory = '/resources/input'
        self.output_directory = '/resources/output'


    def run(self, file_name: str):
        self.comparator = get_comparator()
        self.processing_video(file_name)


    def processing_video(self, file_name: str):
        local_path_to_video = f"{self.input_directory}/{file_name}"
        print(f"File {local_path_to_video} is found: {os.path.isfile(local_path_to_video)}")

        shutil.rmtree(self.output_directory, ignore_errors=True)
        os.makedirs(self.output_directory)

        cap = cv2.VideoCapture(local_path_to_video)

        last_image_for_comparison = None
        print("Starting cap")
        cap_count = 0
        while cap.isOpened():
            ret, frame = cap.read()
            if not ret:
                print('Not ret')
                break
            print(f"Cap count {cap_count}")
            cap_count+=1

            if last_image_for_comparison is None:
                last_image_for_comparison = frame
            else:
                cosine_similarity = self._get_cosine_similarity(last_image_for_comparison, frame)
                if cosine_similarity >= self._similarity_threshold:
                    continue
                last_image_for_comparison = frame

            if self._is_blurred_image(frame):
                continue

            pil_image = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            pil_image.save(os.path.join(self.output_directory, f'{cap_count}.png'), format="PNG")

        cap.release()

        return "OK"




    def _get_cosine_similarity(self, first_frame: NDArray[np.int32], second_frame: NDArray[np.int32]):
        gray_last_image_for_comparison = cvtColor(first_frame, COLOR_BGR2GRAY)
        gray_other_image = cvtColor(second_frame, COLOR_BGR2GRAY)
        gray_last_image_for_comparison = Image.fromarray(gray_last_image_for_comparison)
        gray_other_image = Image.fromarray(gray_other_image)

        comparator = self.comparator

        first_embedding = comparator.encode(gray_last_image_for_comparison, convert_to_tensor=True)
        second_embedding = comparator.encode(gray_other_image, convert_to_tensor=True)

        return util.cos_sim(first_embedding, second_embedding)



    def _is_blurred_image(self, image: NDArray[np.int32]):
        is_blur, var = self._is_blurry(image, self._blur_threshold)
        if is_blur:
            return True
        return False

    def _is_blurry(self, image: NDArray[np.int32], threshold: int = 100):
        laplacian_var = Laplacian(image, CV_64F).var()
        return laplacian_var < threshold, laplacian_var


from unittest import TestCase
from ...framework import Loc, FileIO, LocalExecutor
from pathlib import Path

def check_if_its_sound(content, tc: TestCase):
    import soundfile as sf
    with Loc.create_test_file() as tmp_filename:
        FileIO.write_bytes(content, tmp_filename)
        f = sf.SoundFile(tmp_filename)
        duration = f.frames / f.samplerate
        tc.assertGreater(duration, 1)
    f.close()


VOICEOVER_TEXT = 'The quick brown fox jumps over the lazy dog'

def download_file(url: str, file: Path|str):
    LocalExecutor().execute(['curl', '-L', url, '--output', str(file)])


from pydantic import BaseModel
from typing import Dict, List, Any

class CoordsFaceSquare(BaseModel):
    faces_coords: Dict[str, Dict[str, int]]

class ImageFormat(BaseModel):
    image_base64: str

class LoadModelRequest(BaseModel):
    model_id: str



import json

def button(address, title, parameters, control_parameters: dict[str,str]|None = None, active=True):
    btn = f'<button '
    if not active:
        btn += 'disabled '  # Add the 'disabled' attribute if the button is inactive
    btn+=f'''onClick="makeRequest('{address}','''
    btn += json.dumps(parameters).replace('"', "'")
    btn += ','
    if control_parameters is None:
        btn+='{}'
    else:
        btn+=json.dumps(control_parameters).replace('"', "'")
    btn += f')">{title}</button>'
    return btn

def page(html):
    return start + '\n'.join(html) + end


start = '''
<!DOCTYPE html>
<html lang="en">
<head>
    <title>Brainbox</title>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    </style>
</head>
<body>
'''

end = '''
    <script>
        function makeRequest(address, jsonParameters, inputs) {
            console.log("Sending")
            // Disable all buttons on the page
            const buttons = document.querySelectorAll('button');
            buttons.forEach(button => button.disabled = true);
            
            for (const [key, elementId] of Object.entries(inputs)) {
                const inputElement = document.getElementById(elementId);
                if (inputElement) {
                    jsonParameters[key] = inputElement.value;
                }
            }
            
            console.log(jsonParameters)

            // Send POST request
            fetch(address, {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json'
                },
                body: JSON.stringify({"arguments":jsonParameters})
            })
            .then(response => {
                //buttons.forEach(button => button.disabled = false);
                window.location.reload();
            })
            .catch(error => {
                console.error('There was a problem with the fetch operation:', error);
                // Optionally re-enable buttons if there is an error
                buttons.forEach(button => button.disabled = false);
            });
        }
    </script>
</body>
</html>
    '''


from ...framework import IDecider
import string
import random

class FakeText(IDecider):
    def __call__(self, prefix = '', length = 50):
        characters = string.ascii_letters + string.digits + string.punctuation
        random_text = prefix+' '+''.join(random.choices(characters, k=length))
        return random_text



from ....framework import IDecider
from typing import Generic, TypeVar

TSettings = TypeVar('TSettings')
TController = TypeVar('TController')

class OnDemandDockerApi(IDecider, Generic[TSettings, TController]):
    @property
    def controller(self) -> TController:
        return self._controller

    @controller.setter
    def controller(self, controller: TController):
        self._controller = controller


class HTML:
    @staticmethod
    def escape_code(code):
        code = (
            code
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace('&', "&amp;")
            .replace(" ", "&nbsp")
            .replace("\n", "<br>")
        )
        return code


from typing import *
from yo_fluq import Queryable
from .task_builder import TaskBuilder, IBrainBoxTask

class FunctionalTaskBuilder:
    def __init__(self, *selectors: Callable, method: str='to_media_library'):
        self.selectors = selectors
        self.method = method

    def __call__(self, queryable: Queryable):
        builder = TaskBuilder()
        for tags in queryable:
            task = tags
            for selector in self.selectors:
                task = selector(task)
            if not isinstance(task, IBrainBoxTask):
                raise ValueError(f"After all selectors, BrainBoxTask was expected, but was {task}")
            builder.append(task, tags)
        return builder.to_collector_pack(self.method)


from ..app import BrainBoxApi, IBrainBoxTask
from typing import TypeVar, Generic

T = TypeVar("T")

class BrainBoxCommand(Generic[T]):
    def __init__(self, task: IBrainBoxTask):
        self.task = task

    def execute(self, api: BrainBoxApi) -> T:
        return api.execute(self.task)

    def add(self, api: BrainBoxApi):
        api.add(self.task)

    def join(self, api: BrainBoxApi) -> T:
        return api.join(self.task)

